import logging
import sys

from procmux.args import cli_args
from procmux.config import ProcMuxConfig, parse_config
from procmux.log import formatter, logger
from procmux.server.client import SignalClient


def run_app(cfg: ProcMuxConfig):
    if cfg.log_file:
        file_handler = logging.FileHandler(cfg.log_file)
        file_handler.setLevel(logging.DEBUG)
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
    from procmux.tui.app import start_tui

    start_tui(cfg)


def start_cli():
    config = parse_config(cli_args.config, cli_args.config_override)
    if cli_args.subcommand == "start":
        run_app(config)
    else:
        try:
            signal_client = SignalClient(config)
            if cli_args.subcommand == "signal-start":
                name = cli_args.name
                signal_client.start_process(name)
            elif cli_args.subcommand == "signal-stop":
                name = cli_args.name
                signal_client.stop_process(name)
            elif cli_args.subcommand == "signal-restart":
                name = cli_args.name
                signal_client.restart_process(name)
            elif cli_args.subcommand == "signal-restart-running":
                signal_client.restart_running_processes()
            elif cli_args.subcommand == "signal-stop-running":
                signal_client.stop_running_processes()
        except ValueError as e:
            print(e)
            sys.exit(1)



from typing import Callable, List, Optional
from prompt_toolkit.application import get_app
from prompt_toolkit.layout import Dimension, Float, FloatContainer, HSplit, VSplit
from prompt_toolkit.widgets import Box, Button, Frame, TextArea

from procmux.config import ProcMuxConfig
from procmux.log import logger
from procmux.tui.keybindings import DocumentedKeybindings
from procmux.tui.types import Process
from procmux.util.interpolation import Interpolation


class InterpolationDialog():

    def __init__(
        self, process: Process, run_in_background: bool, config: ProcMuxConfig,
        float_container: FloatContainer,
        on_finished: Callable[[Process, bool, Optional[List[Interpolation]]],
                              None]):
        self._process = process
        self._run_in_background = run_in_background
        self._interpolations = process.config.interpolations
        self._config = config
        self._float_container = float_container
        self._on_finished = on_finished
        self._tab_ix = 0

        self._text_inputs = [
            TextArea(
                height=1,
                width=50,
                prompt=self._config.layout.field_replacement_prompt.replace(
                    '__FIELD_NAME__', interp.field),
                style="class:input-field",
                multiline=False,
                wrap_lines=False,
                text=interp.default_value,
                focus_on_click=True,
            ) for interp in self._interpolations
        ]
        self._move_cursors_to_last_character()

        start_button = Button('Start', self._on_start)
        cancel_button = Button('Cancel', self._on_cancel)
        self._focusable_components = [
            *self._text_inputs, start_button, cancel_button
        ]

        self._container = HSplit([
            Box(body=HSplit(children=self._text_inputs, padding=1),
                padding=Dimension(min=1, max=1, preferred=1),
                padding_bottom=0),
            Box(body=VSplit(children=[start_button, cancel_button],
                            padding=Dimension(min=1, max=5, preferred=5)),
                height=Dimension(min=1, max=3, preferred=3))
        ])

    def _move_cursors_to_last_character(self):
        for ti in self._text_inputs:
            buff = ti.control.buffer
            buff.cursor_position = len(buff.document.current_line_after_cursor)

    def _on_start(self):
        final_input_interpolations: List[Interpolation] = []
        for input_field, interp in zip(self._text_inputs,
                                       self._interpolations):
            final_input_interpolations.append(
                Interpolation(field=interp.field,
                              value=input_field.text,
                              default_value=interp.default_value))
        self._float_container.floats = []
        self._on_finished(self._process, self._run_in_background,
                          final_input_interpolations)

    def _on_cancel(self):
        self._float_container.floats = []
        self._on_finished(self._process, self._run_in_background, None)

    def _move_focus(self, direction: int):
        self._tab_ix = (self._tab_ix + direction) % len(
            self._focusable_components)
        current_input = self._focusable_components[self._tab_ix]
        app = get_app()
        if app:
            app.layout.focus(current_input)

    def get_keybindings(self) -> DocumentedKeybindings:

        def next_input():
            logger.info(
                'interpolation dialog next_input - focusing on next tab index')
            self._move_focus(1)

        def previous_input():
            logger.info(
                'interpolation dialog previous_input - focusing on next tab index'
            )
            self._move_focus(-1)

        kb = DocumentedKeybindings()
        kb.register_configured_keybinding_sans_event(
            self._config.keybinding.next_input, next_input, 'next input')
        kb.register_configured_keybinding_sans_event(
            self._config.keybinding.previous_input, previous_input,
            'previous input')
        kb.register_configured_keybinding_sans_event(
            self._config.keybinding.submit_dialog, self._on_start, 'start')
        kb.register_configured_keybinding_sans_event(
            self._config.keybinding.cancel_dialog, self._on_cancel, 'cancel')
        return kb

    def start_interpolation(self):
        app = get_app()
        if app and self._text_inputs:
            self._float_container.floats = [
                Float(Frame(self._container, self._process.name))
            ]
            app.layout.focus(self._text_inputs[0])
        else:
            self._on_cancel()



from copy import deepcopy
from functools import cached_property
from typing import Any, Callable, Dict, List, Optional, Union

from prompt_toolkit.application import get_app
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.layout import FloatContainer, Window
from prompt_toolkit.mouse_events import MouseEvent, MouseEventType
from ptterm import Terminal

from procmux.config import ProcMuxConfig
from procmux.log import logger
from procmux.server.server import start_server
from procmux.tui.controller.terminal_controller import TerminalController
from procmux.tui.interpolation_dialog import InterpolationDialog
from procmux.tui.keybindings import DocumentedKeybindings
from procmux.tui.state.process_state import ProcessState
from procmux.tui.state.tui_state import TUIState
from procmux.tui.types import FocusTarget, FocusWidget, Process
from procmux.util.interpolation import Interpolation


class TUIController:

    def __init__(self, config: ProcMuxConfig, terminal_placeholder: Window,
                 float_container: FloatContainer):
        self._terminal_placeholder = terminal_placeholder
        self._float_container = float_container
        self._tui_state: TUIState = TUIState(config)
        self._process_state: ProcessState = ProcessState(config)
        self._terminal_controllers: Dict[int, TerminalController] = \
            self._create_terminal_controllers(config, self._process_state.process_list)
        self._filter_change_handlers: List[Callable[[str], None]] = []

        self._server_controller = None
        if config.signal_server.enable:

            def _start_process_and_refresh(process: Process):
                self.start_process(process)
                self.refresh_app()

            self._server_controller = start_server(config, self._process_state,
                                                   self._terminal_controllers,
                                                   _start_process_and_refresh)

    @property
    def float_container(self) -> FloatContainer:
        return self._float_container

    @cached_property
    def config(self) -> ProcMuxConfig:
        return deepcopy(self._tui_state.config)

    @property
    def zoomed_in(self) -> bool:
        return self._tui_state.zoomed_in

    @property
    def filter_mode(self) -> bool:
        return self._tui_state.filter_mode

    @property
    def docs_open(self) -> bool:
        return self._tui_state.docs_open

    @property
    def quitting(self) -> bool:
        return self._tui_state.quitting

    @property
    def modal_open(self) -> bool:
        return self._tui_state.modal_open

    # Processes

    @property
    def process_list(self) -> List[Process]:
        return self._process_state.process_list

    @property
    def filtered_process_list(self) -> List[Process]:
        return self._process_state.filtered_process_list

    @property
    def selected_process(self) -> Optional[Process]:
        return self._process_state.selected_process

    @property
    def is_selected_process_running(self) -> bool:
        return self._process_state.is_selected_process_running

    def is_selected_process(self, process: Process) -> bool:
        if self.selected_process:
            return self.selected_process.index == process.index
        return False

    def start_process(self, process: Optional[Process] = None):
        logger.info(f'in start_process {process.name if process else ""}')
        if process:
            self.start_process_in_terminal(process)
            return
        if self.selected_process:
            self.start_process_in_terminal(self.selected_process)

    def stop_process(self):
        logger.info('in stop_process')
        if self.current_terminal_controller:
            self.current_terminal_controller.stop_process()

    def on_process_spawned(self, process: Process):
        logger.info(f'in on process spawned: {process.name}')
        process.running = True

    def on_process_done(self, process: Process):
        logger.info(f'in on process done: {process.name}')
        process.running = False
        if self.quitting and not self._process_state.has_running_processes:
            self._quit()
        self.refresh_app()

    # /Processes

    # Terminal

    @property
    def current_terminal_controller(self) -> Optional[TerminalController]:
        if self.selected_process:
            return self._terminal_controllers.get(self.selected_process.index,
                                                  None)
        return None

    @property
    def current_terminal(self) -> Union[Terminal, Window]:
        if self.current_terminal_controller and self.current_terminal_controller.terminal:
            return self.current_terminal_controller.terminal
        return self._terminal_placeholder

    def start_process_in_terminal(self, process: Process):
        logger.info(f'starting {process.name} in terminal')
        if self.quitting:
            return  # if procmux is in the process of quitting, don't start a new process

        terminal_controller = self._terminal_controllers[process.index]
        if terminal_controller:
            run_in_background = self.selected_process is None or self.selected_process.index != process.index
            interpolations: List[Interpolation] = []

            if len(process.config.interpolations) > 0:
                id = InterpolationDialog(process, run_in_background,
                                         self.config, self.float_container,
                                         self.on_finish_interpolation)
                self._tui_state.open_modal(id.get_keybindings())
                id.start_interpolation()
                return

            terminal_controller.spawn_terminal(run_in_background,
                                               interpolations)

    def on_finish_interpolation(self, process: Process,
                                run_in_background: bool,
                                interpolations: Optional[List[Interpolation]]):
        self._tui_state.close_modal()
        self.focus_to_sidebar()
        terminal_controller = self._terminal_controllers[process.index]
        if interpolations and terminal_controller:
            terminal_controller.spawn_terminal(run_in_background,
                                               interpolations)

    # /Terminal

    # Focus

    def _focused_target(self) -> Optional[FocusTarget]:
        if self.modal_open:
            return None
        app = get_app()
        if app:
            for target in self._tui_state.focus_targets:
                if app.layout.has_focus(target.element):
                    return target
        return None

    @property
    def focused_widget(self) -> FocusWidget:
        target = self._focused_target()
        if not target:
            return FocusWidget.TERMINAL
        return target.widget

    @focused_widget.setter
    def focused_widget(self, value: FocusWidget):
        self._set_focus(self._tui_state.get_focus_element(value))

    @property
    def is_focused_on_free_form_input(self):
        return self.focused_widget == FocusWidget.SIDE_BAR_FILTER

    def _set_focus(self, element: Optional[Any]):
        app = get_app()
        if app and element:
            app.layout.focus(element)

    def register_focusable_element(self, widget: FocusWidget, element: Any):
        self._tui_state.register_focusable_element(
            FocusTarget(widget=widget, element=element))

    def deregister_focusable_element(self, widget: FocusWidget):
        self._tui_state.deregister_focusable_element(widget)

    def focus_to_current_terminal(self) -> bool:
        logger.info('focusing on current terminal')
        if isinstance(self.current_terminal, Terminal):
            application = get_app()
            if application:
                logger.info('focusing on selected process terminal')
                application.layout.focus(self.current_terminal)
            return True
        return False

    def toggle_sidebar_terminal_focus(self):
        logger.info(
            f'toggling focus between sidebar and terminal- focus_widget:{self.focused_widget}'
        )
        if self.docs_open:
            logger.info(
                'in toggle_sidebar_terminal_focus - but the docs are open, toggling docs off instead'
            )
            self.toggle_docs()
            return
        if self.zoomed_in:
            logger.info(
                'in toggle_sidebar_terminal_focus - currently zoomed in, toggling zoom off instead'
            )
            self.toggle_zoom()
            return
        if self.focused_widget == FocusWidget.TERMINAL:
            logger.info('currently focused on terminal switching to sidebar')
            self.focus_to_sidebar()
        elif self.focused_widget == FocusWidget.SIDE_BAR_SELECT:
            logger.info('currently focused on sidebar switching to terminal')
            self.focus_to_current_terminal()

    def focus_to_sidebar(self):
        logger.info('focusing on sidebar')
        self.focused_widget = FocusWidget.SIDE_BAR_SELECT

    # /Focus

    # Sidebar

    def on_sidebar_mouse_event(self, event: MouseEvent):
        logger.info(f'in sidebar mouse event: {event}')
        if event.event_type == MouseEventType.MOUSE_UP:
            _, y = event.position
            self._process_state.set_selected_process_by_y_pos(y)
            self.focus_to_sidebar()

    def move_process_selection(self, direction: int):
        if not self.selected_process:
            self._process_state.select_first_process()
            return

        if len(self.filtered_process_list) < 2:
            self._process_state.select_first_process()
            return

        available_indices = [p.index for p in self.filtered_process_list]

        if self.selected_process.index not in available_indices:
            self._process_state.select_first_process()
            return

        current_index = available_indices.index(self.selected_process.index)
        new_index = ((current_index + direction) %
                     len(self.filtered_process_list))
        self._process_state.set_selected_process_by_y_pos(new_index)

    def sidebar_up(self):
        self.move_process_selection(-1)

    def sidebar_down(self):
        self.move_process_selection(1)

    def switch_focus(self):
        logger.info('in switch_focus')
        if not self.is_focused_on_free_form_input:
            self.toggle_sidebar_terminal_focus()

    # /Sidebar

    # Filter

    def register_filter_change_handler(self, handler: Callable[[str], None]):
        self._filter_change_handlers.append(handler)

    def on_filter_change(self):
        for handler in self._filter_change_handlers:
            handler(self._process_state.filter_text)

    def start_filter(self):
        logger.info('in start_filter')
        self._tui_state.filter_mode = True
        self._process_state.apply_filter('')
        self.on_filter_change()
        self.focused_widget = FocusWidget.SIDE_BAR_FILTER

    def update_filter(self, buffer: Buffer):
        logger.info(f'in update filter: {buffer.text}')
        self._process_state.apply_filter(buffer.text)
        self.on_filter_change()

    def close_filter(self):
        logger.info('in close_filter')
        self._tui_state.filter_mode = False
        self.focus_to_sidebar()

    def cancel_filter(self):
        logger.info('in cancel_filter')
        self._process_state.apply_filter('')
        self.on_filter_change()
        self.close_filter()

    # /Filter

    # Docs

    def view_docs(self):
        logger.info('in view_docs')
        self.toggle_docs()

    def close_docs(self):
        logger.info('in close_docs')
        self.toggle_docs()

    def toggle_docs(self):
        logger.info(f'setting docks_open to {not self.docs_open}')
        self._tui_state.docs_open = not self.docs_open
        if self.docs_open:
            self.focused_widget = FocusWidget.DOCS
        else:
            self.focused_widget = FocusWidget.SIDE_BAR_SELECT

    # /Docs

    # Zoom

    def zoom(self):
        logger.info('in zoom')
        if not self.is_focused_on_free_form_input:
            self.toggle_zoom()

    def toggle_zoom(self):
        logger.info('in toggle_zoom')
        if self.zoomed_in:
            logger.info('setting zoomed_in to False')
            self._tui_state.zoomed_in = False
            self.focus_to_sidebar()
        else:
            if self.focus_to_current_terminal():
                logger.info('setting zoomed_in to True')
                self._tui_state.zoomed_in = True

    # /Zoom

    # /Sidebar

    # Scroll

    def on_scroll_mode_change(self, process: Process):
        terminal_controller = self._terminal_controllers[process.index]
        if terminal_controller:
            terminal_controller.on_scroll_mode_change(process.scroll_mode)

    def toggle_scroll(self):
        logger.info('in _toggle_scroll')
        if not self.is_focused_on_free_form_input and self.selected_process:
            self.selected_process.scroll_mode = not self.selected_process.scroll_mode
            self.on_scroll_mode_change(self.selected_process)
            if not self.selected_process.scroll_mode and not self.zoomed_in:
                self.focus_to_sidebar()

    # /Scroll

    # Keybindings

    def get_app_keybindings(self) -> DocumentedKeybindings:
        if self.modal_open:
            return self._tui_state.modal_keybindings

        kb = DocumentedKeybindings()
        if self.focused_widget == FocusWidget.DOCS:
            self._add_docs_keybindings(kb)
        if self.focused_widget == FocusWidget.SIDE_BAR_FILTER:
            self._add_filter_keybindings(kb)
        if self.focused_widget == FocusWidget.SIDE_BAR_SELECT:
            self._add_side_bar_keybindings(kb)
            self._add_terminal_keybindings(kb)
        if self.focused_widget == FocusWidget.TERMINAL:
            self._add_terminal_keybindings(kb)
        return kb

    def _add_docs_keybindings(self, kb: DocumentedKeybindings):
        kb.register_configured_keybinding_sans_event(
            self.config.keybinding.docs, self.close_docs, 'close docs')

    def _add_filter_keybindings(self, kb: DocumentedKeybindings):
        kb.register_configured_keybinding_sans_event(
            self.config.keybinding.filter, self.cancel_filter, 'exit filter')
        kb.register_configured_keybinding_sans_event(
            self.config.keybinding.submit_filter, self.close_filter,
            'submit filter')

    def _add_side_bar_keybindings(self, kb: DocumentedKeybindings):
        kb.register_configured_keybinding_sans_event(self.config.keybinding.up,
                                                     self.sidebar_up, 'up')
        kb.register_configured_keybinding_sans_event(
            self.config.keybinding.down, self.sidebar_down, 'down')

        if self.selected_process and self.selected_process.running:
            kb.register_configured_keybinding_sans_event(
                self.config.keybinding.stop, self.stop_process, 'stop')
        else:
            kb.register_configured_keybinding_sans_event(
                self.config.keybinding.start, self.start_process, 'start')

        kb.register_configured_keybinding_sans_event(
            self.config.keybinding.quit, self.quit, 'quit')

        kb.register_configured_keybinding_sans_event(
            self.config.keybinding.filter, self.start_filter, 'filter')

        kb.register_configured_keybinding_sans_event(
            self.config.keybinding.docs, self.view_docs, 'docs')

        return kb

    def _add_terminal_keybindings(self, kb: DocumentedKeybindings):
        if isinstance(self.current_terminal, Terminal):
            kb.register_configured_keybinding_sans_event(
                self.config.keybinding.switch_focus, self.switch_focus,
                'switch focus')
            kb.register_configured_keybinding_sans_event(
                self.config.keybinding.zoom, self.zoom, 'zoom')
            kb.register_configured_keybinding_sans_event(
                self.config.keybinding.toggle_scroll, self.toggle_scroll,
                'toggle scroll')

    # /Keybindings

    def _create_terminal_controllers(
        self,
        config: ProcMuxConfig,
        process_list: List[Process],
    ) -> Dict[int, TerminalController]:
        return {
            p.index: TerminalController(self, config, p)
            for p in process_list
        }

    def autostart(self):
        logger.info('in autostart')
        for process in self.process_list:
            if process.config.autostart:
                logger.info(f'autostarting {process.name}')
                assert not process.config.interpolations, \
                    'processes with autostart enabled must not have interpolations/field replacements'
                self.start_process(process)

    def quit(self):
        logger.info('in quit')
        application = get_app()

        if not application or self.quitting:
            return  # avoid registering process done handler multiple times
        self._tui_state.quitting = True

        # kill the signal server if it's running
        if self._server_controller:
            self._server_controller.stop()

        logger.info('quit - sending kill signals')
        for tc in self._terminal_controllers.values():
            tc.stop_process()
        if not self._process_state.has_running_processes:
            application.exit()

    def _quit(self):
        application = get_app()
        if application:
            application.exit()

    def refresh_app(self):
        app = get_app()
        if app:
            app.invalidate()



from typing import Any, List, Optional

from procmux.config import ProcMuxConfig
from procmux.tui.keybindings import DocumentedKeybindings
from procmux.tui.types import FocusTarget, FocusWidget


class TUIState:

    def __init__(self, config: ProcMuxConfig):
        self.config: ProcMuxConfig = config
        self.zoomed_in: bool = False
        self.filter_mode: bool = False
        self.docs_open: bool = False
        self.quitting: bool = False
        self._focus_targets: List[FocusTarget] = []
        self._modal_open: bool = False
        self._modal_keybindings: DocumentedKeybindings = DocumentedKeybindings(
        )

    @property
    def focus_targets(self) -> List[FocusTarget]:
        return self._focus_targets

    @property
    def modal_open(self) -> bool:
        return self._modal_open

    @property
    def modal_keybindings(self) -> DocumentedKeybindings:
        return self._modal_keybindings

    def get_focus_element(self, widget: FocusWidget) -> Optional[Any]:
        return next(
            (ft.element for ft in self._focus_targets if ft.widget == widget),
            None)

    def register_focusable_element(self, target: FocusTarget):
        self._focus_targets.append(target)

    def deregister_focusable_element(self, widget: FocusWidget):
        self._focus_targets = [
            fe for fe in self._focus_targets if fe.widget != widget
        ]

    def open_modal(self, kb: DocumentedKeybindings = DocumentedKeybindings()):
        self._modal_open = True
        self._modal_keybindings = kb

    def close_modal(self):
        self._modal_open = False
        self._modal_keybindings = DocumentedKeybindings()



from __future__ import unicode_literals

from typing import Any, List

from prompt_toolkit.buffer import Buffer
from prompt_toolkit.layout import DynamicContainer, HSplit, ScrollbarMargin, Window
from prompt_toolkit.layout.controls import BufferControl, FormattedTextControl
from prompt_toolkit.layout.dimension import D, Dimension
from prompt_toolkit.widgets import Box, Frame

from procmux.tui.controller.tui_controller import TUIController
from procmux.tui.types import FocusWidget


class SideBar:
    _right_padding: int = 5

    def __init__(self, controller: TUIController):
        self._controller: TUIController = controller
        self._fixed_width: int = self._controller.config.layout.processes_list_width

        self._filter_buffer: Buffer = Buffer(
            on_text_changed=self._controller.update_filter
        )
        self._buffer_control: BufferControl = BufferControl(buffer=self._filter_buffer)
        self._list_control: FormattedTextControl = FormattedTextControl(
            text=self._get_text_fragments, show_cursor=False, focusable=True
        )

        self._controller.register_focusable_element(
            FocusWidget.SIDE_BAR_FILTER, self._buffer_control
        )
        self._controller.register_focusable_element(
            FocusWidget.SIDE_BAR_SELECT, self._list_control
        )
        self._controller.register_filter_change_handler(self.on_filter_change)

        def _get_filter_container() -> Window:
            if self._controller.filter_mode or self._filter_buffer.text:
                return Window(
                    content=self._buffer_control,
                    height=D(min=1, max=1),
                    width=D(max=self._fixed_width - 1, weight=1),
                )
            return Window(
                height=D(min=0, max=0), width=D(max=self._fixed_width - 1, weight=1)
            )

        right_margins = []
        if self._controller.config.style.show_scrollbar:
            right_margins.append(ScrollbarMargin(display_arrows=True))

        def _get_sidebar_container():
            width = Dimension(min=self._fixed_width, max=self._fixed_width, weight=1)
            if (
                self._controller.focused_widget
                not in [
                    FocusWidget.SIDE_BAR_FILTER,
                    FocusWidget.SIDE_BAR_SELECT,
                ]
                and self._controller.config.layout.hide_sidebar_when_not_focused
            ):
                # hidden sidebar
                width = Dimension(min=0, max=0, weight=0)
            if self._controller.config.style.show_borders:
                return Frame(
                    title="Processes",
                    style="class:sidebar",
                    width=width,
                    body=sidebar_body,
                )
            else:
                return Box(style="class:sidebar", width=width, body=sidebar_body)

        sidebar_body = HSplit(
            [
                DynamicContainer(get_container=_get_filter_container),
                Window(
                    content=self._list_control,
                    style="class:select-box",
                    height=Dimension(min=5, preferred=100000),
                    cursorline=True,
                    right_margins=right_margins,
                ),
            ]
        )

        self._container = DynamicContainer(get_container=_get_sidebar_container)

    def on_filter_change(self, filter_text: str):
        if filter_text != self._filter_buffer.text:
            self._filter_buffer.text = filter_text

    def _get_text_fragments(self) -> List[Any]:
        result = []
        for process in self._controller.filtered_process_list:
            status = "UP" if process.running else "DOWN"
            status_fg = (
                self._controller.config.style.status_running_color
                if process.running
                else self._controller.config.style.status_stopped_color
            )

            target_width = self._fixed_width - self._right_padding - len(status)
            name = process.name
            if len(name) > target_width:
                name = name[0 : target_width - 3] + "..."
            name_fixed = f"%{target_width * -1}s" % name

            fg_color = self._controller.config.style.unselected_process_color
            bg_color = ""
            pointer_char = " "

            if self._controller.is_selected_process(process):
                result.append(("[SetCursorPosition]", ""))
                bg_color = self._controller.config.style.selected_process_bg_color
                fg_color = self._controller.config.style.selected_process_color
                pointer_char = self._controller.config.style.pointer_char

            result.append(
                (
                    f"fg:{fg_color} bg:{bg_color} bold",
                    f"{pointer_char}{name_fixed}",
                    self._controller.on_sidebar_mouse_event,
                )
            )
            result.append(
                (
                    f"fg:{status_fg} bg:{bg_color} bold",
                    status,
                    self._controller.on_sidebar_mouse_event,
                )
            )
            result.append(("", "\n", self._controller.on_sidebar_mouse_event))

        return result

    def __pt_container__(self):
        return self._container



from typing import Any, Callable, List, Optional

from prompt_toolkit.key_binding import KeyBindings

from procmux.tui.types import KeybindingDocumentation


class DocumentedKeybindings(KeyBindings):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.help_docs: List[KeybindingDocumentation] = []

    def add_keybinding_help_doc(self, label: str, keys: List[str]):
        if keys:
            self.help_docs.append(
                KeybindingDocumentation(label=label, help=keys[0]))

    def register_configured_keybinding_sans_event(
        self,
        keys: List[str],
        handler: Callable[[], None],
        help_label: Optional[str] = None,
    ):
        self.register_configured_keybinding(keys,
                                            lambda _: handler(),
                                            help_label=help_label)

    def register_configured_keybinding(self,
                                       keys: List[str],
                                       handler: Callable[[Any], None],
                                       help_label: Optional[str] = None):
        if 'disabled' in keys:
            return

        for keybinding in keys:

            @self.add(keybinding)
            def _(event) -> None:
                handler(event)

        if help_label:
            self.add_keybinding_help_doc(label=help_label, keys=keys)



import argparse
import os
import sys

parser = argparse.ArgumentParser(description="procmux")

sub_parsers = parser.add_subparsers(help='sub-command help', dest='subcommand')
parser_start = sub_parsers.add_parser('start', help='start procmux')
parser_start.add_argument('--config', required=False)
parser_start.add_argument('--config-override', required=False)

parser_signal_start = sub_parsers.add_parser(
    'signal-start',
    help=
    'send a start signal to processes managed by a running procmux instance')
parser_signal_start.add_argument('--name',
                                 type=str,
                                 help='the process name to send the signal to',
                                 required=True)
parser_signal_start.add_argument('--config', required=False)
parser_signal_start.add_argument('--config-override', required=False)

parser_signal_stop = sub_parsers.add_parser(
    'signal-stop',
    help='send a stop signal to processes managed by a running procmux instance'
)
parser_signal_stop.add_argument('--name',
                                type=str,
                                help='the process name to send the signal to',
                                required=True)
parser_signal_stop.add_argument('--config', required=False)
parser_signal_stop.add_argument('--config-override', required=False)

parser_signal_restart = sub_parsers.add_parser(
    'signal-restart',
    help=
    'send a restart signal to processes managed by a running procmux instance')
parser_signal_restart.add_argument(
    '--name',
    type=str,
    help='the process name to send the signal to',
    required=True)
parser_signal_restart.add_argument('--config', required=False)
parser_signal_restart.add_argument('--config-override', required=False)

parser_signal_restart_running = sub_parsers.add_parser(
    'signal-restart-running',
    help=
    'send a restart signal to all running processes managed by a running procmux instance'
)
parser_signal_restart_running.add_argument('--config', required=False)
parser_signal_restart_running.add_argument('--config-override', required=False)

parser_signal_stop_running = sub_parsers.add_parser(
    'signal-stop-running',
    help=
    'send a stop signal to all running processes managed by a running procmux instance'
)
parser_signal_stop_running.add_argument('--config', required=False)
parser_signal_stop_running.add_argument('--config-override', required=False)

if len(sys.argv) == 1 or sys.argv[1] not in [
        'start', 'signal-start', 'signal-stop', 'signal-restart',
        'signal-restart-running', 'signal-stop-running'
]:
    sys.argv.insert(1, 'start')
    cli_args = parser.parse_args(sys.argv[1:])
else:
    cli_args = parser.parse_args()

if not cli_args.config:
    potential_defaults = [
        'procmux-config.yml', 'procmux-config.yaml', 'procmux.yml',
        'procmux.yaml'
    ]
    for potential_default in potential_defaults:
        if os.path.exists(potential_default):
            cli_args.config = potential_default
            break
    if not cli_args.config:
        raise RuntimeError(
            f"{', '.join(potential_defaults)}  config files were not found in the current directory, please use --config to pass a procmux configuration file."
        )



from dataclasses import dataclass
from enum import Enum, auto
from typing import Any

from procmux.config import ProcessConfig


class FocusWidget(Enum):
    SIDE_BAR_FILTER = auto()
    SIDE_BAR_SELECT = auto()
    TERMINAL = auto()
    TERMINAL_COMMAND_INPUTS = auto()
    DOCS = auto()


@dataclass(frozen=True)
class FocusTarget:
    widget: FocusWidget
    element: Any


@dataclass(frozen=True)
class KeybindingDocumentation:
    label: str
    help: str


@dataclass
class Process:
    index: int
    config: ProcessConfig
    name: str
    running: bool = False
    scroll_mode: bool = False



from typing import Optional

from ptterm import Terminal

from procmux.tui.types import Process


class TerminalState:

    def __init__(self, process: Process):
        self.process = process
        self.running = False
        self.scroll_mode = False
        self.terminal: Optional[Terminal] = None

    @property
    def is_running(self) -> bool:
        return self.terminal is not None and self.running



import os
import signal
import sys
from typing import List, Optional, TYPE_CHECKING

from ptterm import Terminal

from procmux.config import ProcMuxConfig
from procmux.log import logger
from procmux.tui.state.terminal_state import TerminalState
from procmux.tui.types import Process
from procmux.util.interpolation import interpolate, Interpolation

if TYPE_CHECKING:
    from procmux.tui.controller.tui_controller import TUIController


class TerminalController:

    def __init__(self, controller: 'TUIController', config: ProcMuxConfig,
                 process: Process):
        self._controller: TUIController = controller
        self._config: ProcMuxConfig = config
        self._terminal_state = TerminalState(process)

    @property
    def terminal(self) -> Optional[Terminal]:
        return self._terminal_state.terminal

    @property
    def is_running(self) -> bool:
        return self._terminal_state.is_running

    @property
    def _process(self) -> Process:
        return self._terminal_state.process

    def _export_env_vars(self) -> bool:
        if self._process.config.env:
            for key, value in self._process.config.env.items():
                os.environ[key] = value or ''
            return True
        return False

    def _adjust_path(self):
        proc_config = self._process.config
        if proc_config.add_path:
            if type(proc_config.add_path) == str:
                paths = [proc_config.add_path]
            paths = proc_config.add_path
            for p in paths:
                sys.path.append(p)

    def _change_working_directory(self):
        os.chdir(self._process.config.cwd)

    def _get_cmd(
            self,
            interpolations: Optional[List[Interpolation]] = None) -> List[str]:
        proc_config = self._process.config
        cmd = []
        if proc_config.shell:
            cmd.extend(self._config.shell_cmd)
            cmd.append(interpolate(proc_config.shell, interpolations))
        elif proc_config.cmd:
            cmd = [interpolate(c, interpolations) for c in proc_config.cmd]
        return cmd

    def stop_process(self):
        logger.info(f'in top process: {self._process.name}')
        if self.is_running and self.terminal:
            stop_signal = self._process.config.stop
            sig_code = getattr(signal, stop_signal)
            try:
                if hasattr(self.terminal.process.terminal, 'send_signal'):
                    logger.info(
                        f'stopping process {self._process.name} with defined signal {stop_signal}'
                    )
                    self.terminal.process.terminal.send_signal(sig_code)
                else:
                    logger.info(
                        f'killing process {self._process.name} using x-platform process kill()'
                        f' - disregarding defined kill sig')
                    self._kill()
            except Exception as e:
                logger.error(
                    f'failed to kill process: {self._process.name} {e}')

    def _kill(self):
        if self.is_running and self.terminal:
            try:
                logger.info(f'running kill() on process {self._process.name}')
                self.terminal.terminal_control.process.kill()
            except Exception as e:
                logger.error(
                    f'failed to kill process name: {self._process.name} {e}')

    def _handle_process_done(self):
        logger.info(f'{self._process.name} is done')
        self._terminal_state.running = False
        self._controller.on_process_done(self._process)

    def _handle_process_spawned(self):
        logger.info(
            f'created terminal {self.terminal} for process {self._process.name}'
        )
        self._terminal_state.running = True
        self._controller.on_process_spawned(self._process)

    def spawn_terminal(self,
                       run_in_background: bool,
                       interpolations: Optional[List[Interpolation]] = None):
        logger.info(f'in spawn terminal for process {self._process.name}')
        if self.is_running:
            logger.info(
                f'{self._process.name} terminal already running - returning existing terminal - {self.terminal}'
            )
            return

        def before_exec():
            self._change_working_directory()
            self._export_env_vars()
            self._adjust_path()

        self._terminal_state.terminal = Terminal(
            command=self._get_cmd(interpolations),
            width=self._config.style.width_100,
            height=self._config.style.height_100,
            style='class:terminal',
            before_exec_func=before_exec,
            done_callback=self._handle_process_done)
        if run_in_background:
            logger.info(
                f'rendering ptterm in the background, because {self._process.name} is not actively selected'
            )
            if self.terminal:
                self.terminal.terminal_control.create_content(width=100,
                                                              height=100)
        self._handle_process_spawned()

    def on_scroll_mode_change(self, scroll_mode: bool):
        if self.terminal and self._terminal_state.scroll_mode != scroll_mode:
            self._terminal_state.scroll_mode = scroll_mode
            if self._terminal_state.scroll_mode:
                self.terminal.enter_copy_mode()
            else:
                self.terminal.exit_copy_mode()



from __future__ import unicode_literals

from typing import Callable

from prompt_toolkit.formatted_text import FormattedText, HTML, merge_formatted_text
from prompt_toolkit.layout import Window
from prompt_toolkit.layout.controls import FormattedTextControl

from procmux.tui.types import KeybindingDocumentation
from procmux.tui.controller.tui_controller import TUIController


class HelpPanel:

    def __init__(self, controller: TUIController):
        self._controller: TUIController = controller
        self._container: Window = Window(height=1,
                                         content=FormattedTextControl(
                                             text=self._get_formatted_text,
                                             focusable=False,
                                             show_cursor=False))

    def _get_formatted_text(self) -> Callable[[], FormattedText]:
        delimiter = " | "

        def intersperse(lst, item):
            full_list = [item] * (len(lst) * 2 - 1)
            full_list[0::2] = lst
            return full_list

        result = [
            self._get_key_combo_text(help_)
            for help_ in self._controller.get_app_keybindings().help_docs
        ]
        if result:
            result = intersperse(result, delimiter)
        return merge_formatted_text(result)

    def _get_key_combo_text(self, help_doc: KeybindingDocumentation) -> HTML:
        return HTML(f'<b>&lt;{help_doc.help}&gt;</b> {help_doc.label}')

    def __pt_container__(self):
        return self._container



from collections import namedtuple
import re
from typing import List, Optional

_replacement_regex = r"<([^:>]*):?([^>]*)>"

Interpolation = namedtuple('Interpolation',
                           ['field', 'value', 'default_value'])


def parse_interpolations(*cmds) -> List[Interpolation]:
    interpolations = []
    field_names = set()
    for cmd in cmds:
        matches = re.findall(_replacement_regex, cmd)
        for m in matches:
            value = ''
            if m:
                field = m[0].strip()
                if len(m) == 2:
                    value = m[1].strip()
                interpolations.append(
                    Interpolation(field=field,
                                  value=value,
                                  default_value=value))
                assert field not in field_names, f'A duplicate replacement ' \
                                                 f'field detected during interpolation - field: "{field}" \n' \
                                                 f'All replacement fields must have unique names per process definition'
                field_names.add(field)
    return interpolations


def interpolate(text: str,
                interpolations: Optional[List[Interpolation]] = None) -> str:
    if not interpolations:
        return text
    for interp in interpolations:
        replace_regex = rf'<\s*{interp.field}\s*:?\s*{interp.default_value}\s*>'
        text = re.sub(replace_regex, interp.value, text)

    return text



from prompt_toolkit.layout import Dimension, DynamicContainer, VSplit
from prompt_toolkit.widgets import Box, Frame

from procmux.tui.controller.tui_controller import TUIController
from procmux.tui.types import FocusWidget


class TerminalPanel:

    def __init__(self, controller: TUIController):
        self._controller: TUIController = controller
        width = Dimension(min=0, max=100000, weight=1)
        height = Dimension(min=0, max=100000, weight=1)

        body = VSplit([
            DynamicContainer(
                get_container=lambda: self._controller.current_terminal),
        ])
        if self._controller.config.style.show_borders:
            self._container = Frame(title='Terminal',
                                    width=width,
                                    height=height,
                                    body=body)
        else:
            self._container = Box(
                body=body,
                width=width,
                height=height,
            )
        self._controller.register_focusable_element(FocusWidget.TERMINAL,
                                                    element=self)

    def __pt_container__(self):
        return self._container



import http.server
import json
import socketserver
import threading
from http import HTTPStatus
from time import sleep, time
from typing import Any, Callable, Dict, Optional, Tuple, Union
from urllib.parse import unquote

from procmux.config import ProcMuxConfig
from procmux.log import logger
from procmux.tui.controller.terminal_controller import TerminalController
from procmux.tui.state.process_state import ProcessState
from procmux.tui.types import Process

# Timeout (in seconds) for restarting a process
# maybe this can come from a query parameter later on
timeout = 5


def start_server(
    cfg: ProcMuxConfig,
    process_state: ProcessState,
    terminal_controllers: Dict[int, TerminalController],
    start_process_callback: Callable[[Process], None],
):

    active_httpd = None

    def _start_server():

        class SignalServer(http.server.SimpleHTTPRequestHandler):

            def __init__(self, request: bytes, client_address: Tuple[str, int],
                         server: socketserver.BaseServer):
                super().__init__(request, client_address, server)

            def log_error(self, _format: str, *args: Any) -> None:
                logger.error(*args)
                pass

            def log_request(self,
                            code: Union[str, int] = "-",
                            size: Union[str, int] = "-") -> None:
                logger.info(f"Request: {self.path} {code} {size}")

            def log_message(self, _format: str, *args: Any) -> None:
                logger.info.format(*args)

            def _send_ok(self, body: bytes):
                self.send_response(HTTPStatus.OK)
                self.send_header("Content-Type", "application/json")
                self.end_headers()
                self.wfile.write(body)

            def _send_error(self, code: int, message: str):
                self.send_response(code)
                self.send_header("Content-Type", "application/json")
                self.end_headers()
                self.wfile.write(json.dumps({"error": message}).encode())

            def _identify_process_by_name(self) -> Optional[Process]:
                name = self.path.split('/')[-1]
                name = unquote(name)
                for process in process_state.process_list:
                    if process.name == name:
                        return process
                return None

            def handle_get_process_list(self):
                process_list = [{
                    "name": p.name,
                    "running": p.running,
                    "index": p.index,
                    "scroll_mode": p.scroll_mode
                } for p in process_state.process_list]
                resp = json.dumps({"process_list": process_list}).encode()
                self._send_ok(bytes(resp))

            def handle_stop_by_name(self):
                process = self._identify_process_by_name()
                if process:
                    terminal_controller = terminal_controllers.get(
                        process.index)
                    if terminal_controller:
                        terminal_controller.stop_process()
                        self._send_ok(b'{}')
                        return
                self._send_error(HTTPStatus.NOT_FOUND, "Process not found")

            def handle_start_by_name(self):
                process = self._identify_process_by_name()
                if process:
                    if process.config.interpolations:
                        self._send_error(
                            HTTPStatus.BAD_REQUEST,
                            'Process requires interpolations, it cannot be remotely signaled to start'
                        )
                        return

                    start_process_callback(process)
                    self._send_ok(b'{}')
                    return
                self._send_error(HTTPStatus.NOT_FOUND, "Process not found")

            def _wait_for_process_stop(
                    self, terminal_controller: TerminalController):
                start_time = time()
                while terminal_controller.is_running:
                    if time() - start_time > timeout:
                        raise TimeoutError("Failed to stop process")
                    sleep(0.1)

            def handle_restart_by_name(self):
                process = self._identify_process_by_name()
                if process:
                    if process.config.interpolations:
                        self._send_error(
                            HTTPStatus.BAD_REQUEST,
                            'Process requires interpolations, it cannot be remotely signaled to start'
                        )

                    terminal_controller = terminal_controllers.get(
                        process.index)
                    if terminal_controller:
                        terminal_controller.stop_process()
                        try:
                            self._wait_for_process_stop(terminal_controller)
                        except TimeoutError:
                            self._send_error(HTTPStatus.INTERNAL_SERVER_ERROR,
                                             "Failed to stop process")
                            return
                        start_process_callback(process)
                        self._send_ok(b'{}')
                        return
                self._send_error(HTTPStatus.NOT_FOUND, "Process not found")

            def handle_restart_running(self):
                running_processes = [
                    p for p in process_state.process_list
                    if p.running and not p.config.interpolations
                ]
                for process in running_processes:
                    if process.running:
                        terminal_controller = terminal_controllers.get(
                            process.index)
                        if terminal_controller:
                            terminal_controller.stop_process()
                            try:
                                self._wait_for_process_stop(
                                    terminal_controller)
                            except TimeoutError:
                                self._send_error(
                                    HTTPStatus.INTERNAL_SERVER_ERROR,
                                    "Failed to stop process")
                                return
                            start_process_callback(process)
                self._send_ok(b'{}')

            def handle_stop_running(self):
                running_processes = [
                    p for p in process_state.process_list if p.running
                ]
                for process in running_processes:
                    terminal_controller = terminal_controllers.get(
                        process.index)
                    if terminal_controller:
                        terminal_controller.stop_process()
                self._send_ok(b'{}')

            def do_GET(self):
                if self.path == '/':
                    self.handle_get_process_list()
                else:
                    self._send_error(HTTPStatus.NOT_FOUND,
                                     "Endpoint not found")

            def do_POST(self):
                if self.path.startswith('/stop-by-name/'):
                    self.handle_stop_by_name()
                elif self.path.startswith('/start-by-name/'):
                    self.handle_start_by_name()
                elif self.path.startswith('/restart-by-name/'):
                    self.handle_restart_by_name()
                elif self.path.startswith('/restart-running'):
                    self.handle_restart_running()
                elif self.path.startswith('/stop-running'):
                    self.handle_stop_running()
                else:
                    self._send_error(HTTPStatus.NOT_FOUND,
                                     "Endpoint not found")

        with socketserver.TCPServer(
            (cfg.signal_server.host, cfg.signal_server.port),
                SignalServer) as httpd:
            nonlocal active_httpd
            active_httpd = httpd
            httpd.serve_forever()

    class ServerController:

        def __init__(self):
            self.thread = None

        def start(self):
            self.thread = threading.Thread(target=_start_server)
            self.thread.daemon = True
            self.thread.start()

        def stop(self):
            if active_httpd:
                active_httpd.shutdown()
            if self.thread:
                self.thread.join(5)

    controller = ServerController()
    controller.start()

    return controller



from __future__ import unicode_literals

from prompt_toolkit.application import Application
from prompt_toolkit.key_binding import DynamicKeyBindings
from prompt_toolkit.layout import ConditionalContainer, DynamicContainer, FloatContainer, HSplit, Layout, VSplit, Window
from prompt_toolkit.styles import Style

from procmux.config import ProcMuxConfig
from procmux.tui.controller.tui_controller import TUIController
from procmux.tui.view.docs import DocsDialog
from procmux.tui.view.help import HelpPanel
from procmux.tui.view.process_description import ProcessDescriptionPanel
from procmux.tui.view.side_bar import SideBar
from procmux.tui.view.terminal import TerminalPanel


def start_tui(config: ProcMuxConfig):
    terminal_placeholder = Window(
        style=f'bg:{config.style.placeholder_terminal_bg_color}',
        width=config.style.width_100,
        height=config.style.height_100)

    dynamic_container = DynamicContainer(get_container=lambda: None)
    float_container = FloatContainer(content=dynamic_container, floats=[])

    controller = TUIController(config, terminal_placeholder, float_container)

    side_bar = SideBar(controller)

    main_layout_container = HSplit([
        VSplit([side_bar, TerminalPanel(controller)]),
        ConditionalContainer(
            content=ProcessDescriptionPanel(controller),
            filter=not controller.config.layout.hide_process_description_panel
        ),
        ConditionalContainer(content=HelpPanel(controller),
                             filter=not controller.config.layout.hide_help)
    ])

    docs_layout_container = HSplit([
        DocsDialog(controller),
        ConditionalContainer(
            content=ProcessDescriptionPanel(controller),
            filter=not controller.config.layout.hide_process_description_panel
        ),
        ConditionalContainer(content=HelpPanel(controller),
                             filter=not controller.config.layout.hide_help)
    ])

    def _get_layout_container():
        if controller.zoomed_in:
            return controller.current_terminal
        elif controller.docs_open:
            return docs_layout_container
        return main_layout_container

    dynamic_container.get_container = _get_layout_container

    application = Application(
        layout=Layout(container=controller.float_container,
                      focused_element=side_bar),
        full_screen=True,
        mouse_support=controller.config.enable_mouse,
        key_bindings=DynamicKeyBindings(
            get_key_bindings=controller.get_app_keybindings),
        style=Style(list((controller.config.style.style_classes
                          or {}).items())),
        color_depth=controller.config.style.color_depth)

    controller.autostart()
    application.run()



from __future__ import unicode_literals
from typing import Callable

from prompt_toolkit.formatted_text import HTML, merge_formatted_text
from prompt_toolkit.formatted_text.base import FormattedText
from prompt_toolkit.layout import Window
from prompt_toolkit.layout.controls import FormattedTextControl

from procmux.tui.controller.tui_controller import TUIController


class ProcessDescriptionPanel:

    def __init__(self, controller: TUIController):
        self._controller: TUIController = controller
        self._container: Window = Window(height=1,
                                         content=FormattedTextControl(
                                             text=self._get_formatted_text,
                                             focusable=False,
                                             show_cursor=False))

    def _get_formatted_text(self) -> Callable[[], FormattedText]:
        process = self._controller.selected_process
        if not process:
            return merge_formatted_text([HTML('')])
        desc = " - " + process.config.description if process.config.description else ''
        result = [HTML(f'<b>{process.name}</b>{desc}')]
        return merge_formatted_text(result)

    def __pt_container__(self):
        return self._container



import os
from dataclasses import dataclass, field, fields
from typing import Dict, List, Literal, Optional, OrderedDict, Union

import hiyapyco
from prompt_toolkit.layout import D, Dimension
from prompt_toolkit.output import ColorDepth

from procmux.util.interpolation import Interpolation, parse_interpolations


class MisconfigurationError(Exception):
    pass


@dataclass
class ProcessConfig:
    """
    shell:(str) command to run (exactly one of shell or cmd must be provided).
    cmd: (list) Array of command and args to run (exactly one of shell or cmd must be provided).
    cwd:(str) Set working directory for the process.
    env: (Dict[str,str]) - Set env variables. Object keys are variable names.
    add_path: string|array - Add entries to the PATH environment variable.
    autostart: bool - Start process when procmux starts.
    stop: "SIGINT"|"SIGTERM"|"SIGKILL" - default will SIGKILL
    """

    autostart: bool = False
    shell: Optional[str] = None
    cmd: Optional[List[str]] = None
    cwd: str = field(default_factory=lambda: os.getcwd())
    stop: str = "SIGKILL"
    env: Optional[Dict[str, Optional[str]]] = None
    add_path: Optional[Union[str, List[str]]] = None
    description: Optional[str] = None
    docs: Optional[str] = None
    categories: Optional[List[str]] = None
    meta_tags: Optional[List[str]] = None

    def __post_init__(self):
        self.validate()

    def validate(self):
        if not self.cmd and not self.shell:
            raise MisconfigurationError(
                "shell or cmd is required for every proc definition"
            )

    @property
    def interpolations(self) -> List[Interpolation]:
        if self.cmd:
            return parse_interpolations(*self.cmd)
        return parse_interpolations(self.shell)


@dataclass
class LayoutConfig:
    hide_help: bool = False
    hide_process_description_panel: bool = False
    hide_sidebar_when_not_focused: bool = False
    processes_list_width: int = 31
    sort_process_list_alpha: bool = True
    category_search_prefix: str = "cat:"
    field_replacement_prompt: str = "__FIELD_NAME__ ⮕  "


@dataclass
class StyleConfig:
    selected_process_color: str = "ansiblack"
    selected_process_bg_color: str = "ansimagenta"
    unselected_process_color: str = "ansiblue"
    status_running_color: str = "ansigreen"
    status_stopped_color: str = "ansired"
    placeholder_terminal_bg_color: str = "#1a1b26"
    pointer_char: str = "▶"
    show_borders: bool = True
    show_scrollbar: bool = True
    style_classes: Optional[Dict[str, str]] = None
    color_level: Optional[
        Union[
            Literal["monochrome"],
            Literal["ansicolors"],
            Literal["256colors"],
            Literal["truecolors"],
        ]
    ] = None

    @property
    def color_depth(
        self,
    ) -> Union[
        Literal["DEPTH_1_BIT"],
        Literal["DEPTH_4_BIT"],
        Literal["DEPTH_8_BIT"],
        Literal["DEPTH_24_BIT"],
    ]:
        if self.color_level == "monochrome":
            return ColorDepth.MONOCHROME
        if self.color_level == "ansicolors":
            return ColorDepth.ANSI_COLORS_ONLY
        if self.color_level == "256colors":
            return ColorDepth.DEFAULT
        return ColorDepth.TRUE_COLOR

    @property
    def width_100(self) -> Dimension:
        return D(preferred=100 * 100)

    @property
    def height_100(self) -> Dimension:
        return D(preferred=100 * 100)


@dataclass
class KeybindingConfig:
    quit: List[str] = field(default_factory=lambda: ["q"])
    filter: List[str] = field(default_factory=lambda: ["/"])
    submit_filter: List[str] = field(default_factory=lambda: ["enter"])
    next_input: List[str] = field(default_factory=lambda: ["tab", "down"])
    previous_input: List[str] = field(default_factory=lambda: ["s-tab", "up"])
    submit_dialog: List[str] = field(default_factory=lambda: ["enter"])
    cancel_dialog: List[str] = field(default_factory=lambda: ["escape"])
    start: List[str] = field(default_factory=lambda: ["s"])
    stop: List[str] = field(default_factory=lambda: ["x"])
    up: List[str] = field(default_factory=lambda: ["up", "k"])
    down: List[str] = field(default_factory=lambda: ["down", "j"])
    switch_focus: List[str] = field(default_factory=lambda: ["c-w"])
    zoom: List[str] = field(default_factory=lambda: ["c-z"])
    docs: List[str] = field(default_factory=lambda: ["?"])
    toggle_scroll: List[str] = field(default_factory=lambda: ["c-s"])

    def __post_init__(self):
        for keybinding_field in fields(KeybindingConfig):
            keybinding = getattr(self, keybinding_field.name)
            if isinstance(keybinding, str):
                keybinding = [keybinding]
                setattr(self, keybinding_field.name, keybinding)


@dataclass
class SignalServerConfig:
    port: int = 9792
    host: str = "localhost"
    enable: bool = False


@dataclass
class ProcMuxConfig:
    procs: Dict[str, ProcessConfig] = field(default_factory=dict)
    signal_server: SignalServerConfig = field(default_factory=SignalServerConfig)
    style: StyleConfig = field(default_factory=StyleConfig)
    keybinding: KeybindingConfig = field(default_factory=KeybindingConfig)
    shell_cmd: List[str] = field(
        default_factory=lambda: [os.environ.get("SHELL", "/bin/sh"), "-c"]
    )
    layout: LayoutConfig = field(default_factory=LayoutConfig)
    log_file: Optional[str] = None
    enable_mouse: bool = True

    def __post_init__(self):
        def is_dict_like(obj):
            return isinstance(obj, (dict, OrderedDict))

        if is_dict_like(self.procs):
            process_config_data = {}
            for proc_key, proc_value in self.procs.items():
                if isinstance(proc_value, dict):
                    process_config_data[proc_key] = ProcessConfig(**proc_value)
                else:
                    process_config_data[proc_key] = proc_value
            self.procs = process_config_data
        if is_dict_like(self.style):
            self.style = StyleConfig(**self.style)
        if is_dict_like(self.keybinding):
            self.keybinding = KeybindingConfig(**self.keybinding)
        if is_dict_like(self.layout):
            self.layout = LayoutConfig(**self.layout)
        if is_dict_like(self.signal_server):
            self.signal_server = SignalServerConfig(**self.signal_server)


def parse_config(
    config_file: str, override_config_file: Optional[str] = None
) -> ProcMuxConfig:
    config_files = [config_file]
    if override_config_file:
        config_files.append(override_config_file)

    config_dict = hiyapyco.load(
        *config_files, method=hiyapyco.METHOD_SIMPLE, failonmissingfiles=True
    )
    return ProcMuxConfig(**config_dict)



from typing import Dict, List, Optional

from procmux.config import ProcMuxConfig, ProcessConfig
from procmux.tui.types import Process


class ProcessState:

    def __init__(self, config: ProcMuxConfig):
        self.config: ProcMuxConfig = config
        self.process_list: List[Process] = self._create_process_list(
            self.config.procs)
        self.filtered_process_list: List[Process] = self.process_list
        self.selected_process: Optional[Process] = self.filtered_process_list[
            0] if self.filtered_process_list else None
        self._filter: str = ''

    @property
    def is_selected_process_running(self) -> bool:
        return self.selected_process.running if self.selected_process else False

    @property
    def has_running_processes(self) -> bool:
        for process in self.process_list:
            if process.running:
                return True
        return False

    @property
    def filter_text(self) -> str:
        return self._filter

    def select_first_process(self):
        if self.filtered_process_list:
            self.selected_process = self.filtered_process_list[0]

    def set_selected_process_by_y_pos(self, y_pos: int):
        if self.filtered_process_list and y_pos >= 0 and y_pos < len(
                self.filtered_process_list):
            self.selected_process = self.filtered_process_list[y_pos]

    def _create_process_list(
            self, process_config: Dict[str, ProcessConfig]) -> List[Process]:
        return self._sort_process_list([
            Process(ix, pc, n)
            for ix, (n, pc) in enumerate(process_config.items())
        ])

    def _sort_process_list(self, ps: List[Process]) -> List[Process]:
        if self.config.layout.sort_process_list_alpha:
            return sorted(ps, key=lambda p: p.name)
        return ps

    def apply_filter(self, filter_text: str):
        self._filter = filter_text

        if not self._filter:
            self.filtered_process_list = self._sort_process_list(
                self.process_list)

        prefix = self.config.layout.category_search_prefix

        def filter_against_category(search_text: str,
                                    process: Process) -> bool:
            search_t = search_text[len(prefix):]
            if not process.config.categories:
                return False
            categories = {c.lower() for c in process.config.categories}
            return search_t.lower() in categories

        def filter_against_name_and_meta(search_text: str,
                                         process: Process) -> bool:
            tags = set()
            if process.config.meta_tags:
                for tag in process.config.meta_tags:
                    tags.add(tag.lower())
            return search_text.lower() in process.name or search_text.lower(
            ) in tags

        def filter_(search_text: str, process: Process) -> bool:
            if search_text.startswith(prefix):
                return filter_against_category(search_text, process)
            return filter_against_name_and_meta(search_text, process)

        self.filtered_process_list = self._sort_process_list(
            [p for p in self.process_list if filter_(self._filter, p)])
        self.selected_process = self.filtered_process_list[
            0] if self.filtered_process_list else None



from typing import Callable, Union

from prompt_toolkit import HTML
from prompt_toolkit.formatted_text import merge_formatted_text
from prompt_toolkit.formatted_text.base import FormattedText
from prompt_toolkit.layout import FormattedTextControl, Window
from prompt_toolkit.widgets import Frame

from procmux.tui.types import FocusWidget
from procmux.tui.controller.tui_controller import TUIController


class DocsDialog:

    def __init__(self, controller: TUIController):
        self._controller: TUIController = controller
        self._container: Frame = Frame(
            title=self._get_title,
            body=Window(
                content=FormattedTextControl(text=self._get_formatted_text,
                                             focusable=True,
                                             show_cursor=False)))
        self._controller.register_focusable_element(FocusWidget.DOCS,
                                                    self._container)

    def _get_title(self) -> str:
        process = self._controller.selected_process
        return process.name if process else 'Help'

    def _get_formatted_text(self) -> Union[HTML, Callable[[], FormattedText]]:
        process = self._controller.selected_process
        if process:
            result = []
            if process.config.description:
                result.append(HTML(f'<b>{process.config.description}</b>\n'))
            if process.config.docs:
                result.append(HTML(process.config.docs))
            if len(result) == 0:
                result.append(f'No docs available for process: {process.name}')

            return merge_formatted_text(result)
        return HTML('No process is currently selected.')

    def __pt_container__(self):
        return self._container



import http.client
import json
from urllib.parse import quote

from procmux.config import ProcMuxConfig


class SignalClient:

    def __init__(self, config: ProcMuxConfig):
        if not config.signal_server.enable:
            raise ValueError('Signal server is not enabled in config')
        if not config.signal_server.port:
            raise ValueError('Signal server port is not set in config')
        if not config.signal_server.host:
            raise ValueError('Signal server host is not set in config')

        self._base_url = config.signal_server.host
        self._port = config.signal_server.port

    def _get_error_message(self, response: http.client.HTTPResponse) -> str:
        body = response.read()
        body_json = json.loads(body.decode())
        if body_json.get('error'):
            return body_json['error']
        return body.decode()

    def restart_process(self, name: str):
        name = quote(name)
        conn = http.client.HTTPConnection(self._base_url, self._port)
        conn.request("POST", f"/restart-by-name/{name}")
        response = conn.getresponse()
        if response.status != 200:
            raise ValueError(
                f"Failed to restart process: {response.status} {self._get_error_message(response)}"
            )
        conn.close()

    def stop_process(self, name: str):
        name = quote(name)
        conn = http.client.HTTPConnection(self._base_url, self._port)
        conn.request("POST", f"/stop-by-name/{name}")
        response = conn.getresponse()
        if response.status != 200:
            raise ValueError(
                f"Failed to stop process: {response.status} {self._get_error_message(response)}"
            )
        conn.close()

    def restart_running_processes(self):
        conn = http.client.HTTPConnection(self._base_url, self._port)
        conn.request("POST", "/restart-running")
        response = conn.getresponse()
        if response.status != 200:
            raise ValueError(
                f"Failed to restart running processes: {response.status} {self._get_error_message(response)}"
            )
        conn.close()

    def stop_running_processes(self):
        conn = http.client.HTTPConnection(self._base_url, self._port)
        conn.request("POST", "/stop-running")
        response = conn.getresponse()
        if response.status != 200:
            raise ValueError(
                f"Failed to stop running processes: {response.status} {self._get_error_message(response)}"
            )
        conn.close()

    def start_process(self, name: str):
        name = quote(name)
        conn = http.client.HTTPConnection(self._base_url, self._port)
        conn.request("POST", f"/start-by-name/{name}")
        response = conn.getresponse()
        if response.status != 200:
            raise ValueError(
                f"Failed to start process: {response.status}  {self._get_error_message(response)}"
            )
        conn.close()

    def get_process_list(self):
        conn = http.client.HTTPConnection(self._base_url, self._port)
        conn.request("GET", "/")
        response = conn.getresponse()
        if response.status != 200:
            raise ValueError(
                f"Failed to get process list: {response.status} {self._get_error_message(response)}"
            )
        data = response.read()
        conn.close()
        return data
